@@ -1,5 +1,4 @@
-# author: Martin Maly
-# email: martin.maly.mm@email.cz
+# coding: utf-8
 import argparse
 import os
 import sys
@@ -13,12 +12,9 @@ import json
 try:
     from cctbx import miller, crystal, uctbx, sgtbx, xray
     from cctbx.array_family import flex
+    from iotbx import mtz, reflection_file_reader
 except ModuleNotFoundError:
     print("WARNING: ModuleNotFoundError: Module CCTBX was not found.")
-# TO DO: 3 input files
-# TO DO: unit cell parameters from stream file
-# TO DO: check output MTZ
-# TO DO: wavelength
 
 
 def hkl_strip(hklin):
@@ -103,7 +99,7 @@ def calc_CCstar(CChalf):
     return CCstar
 
 
-def get_miller_array(hklin, cs, values="I", d_max=0, d_min=0):
+def get_miller_array_crystfel(hklin, cs, values="I", d_max=0, d_min=0):
     assert values == "I" or values == "nmeas"
     # read data from the text file
     # expected format of a fixed-width .hkl file:
@@ -141,36 +137,38 @@ def get_miller_array(hklin, cs, values="I", d_max=0, d_min=0):
 
 def calc_stats_merged(m_all_i, m_all_nmeas, d_max=0, d_min=0, n_bins=10):
     stats = {"overall": {}, "binned": {}}
-    m_all_i = m_all_i.resolution_filter(d_max=d_max, d_min=d_min)
     res_low, res_high = m_all_i.d_max_min()
-    n_ref = m_all_i.size()
+    n_unique = m_all_i.size()
     completeness = m_all_i.as_non_anomalous_set().completeness()
     m_all_i = m_all_i.map_to_asu()
+    m_all_i.setup_binner(n_bins=n_bins)
     m_all_i = m_all_i.sort("packed_indices")
-    m_all_nmeas = m_all_nmeas.resolution_filter(d_max=d_max, d_min=d_min)
+    m_all_i.setup_binner(n_bins=n_bins) # need to redo the binner as is lost above
 
     # overall values
     i_mean = m_all_i.mean()
     i_sig = m_all_i.i_over_sig_i()
+    n_obs = int(m_all_nmeas.sum())
     multiplicity = m_all_nmeas.mean()
     stats["overall"]["d_max"] = round(res_low, 3)
     stats["overall"]["d_min"] = round(res_high, 3)
-    stats["overall"]["n_unique"] = n_ref
-    stats["overall"]["completeness"] = round(completeness * 100, 1)
-    stats["overall"]["multiplicity"] = round(multiplicity, 1)
+    stats["overall"]["n_unique"] = n_unique
+    stats["overall"]["n_obs"] = n_obs
+    stats["overall"]["completeness"] = round(completeness * 100, 2)
+    stats["overall"]["multiplicity"] = round(multiplicity, 2)
     stats["overall"]["I"] = round(i_mean, 2)
     stats["overall"]["IsigI"] = round(i_sig, 2)
+    print(f"#observed: {n_obs}")
     print(f"#unique: {m_all_i.size()}")
-    print(f"completeness = {completeness * 100:.1f} %")
-    print(f"multiplicity = {multiplicity:.1f}")
-    print(f"<I> = {i_mean:.2f}")
-    print(f"<I/sigma(I)> = {i_sig:.2f}")
+    print(f"completeness = {completeness * 100:.2f} %")
+    print(f"multiplicity = {multiplicity:.2f}")
+    print(f"<I> = {i_mean:.1f}")
+    print(f"<I/sigma(I)> = {i_sig:.1f}")
 
     # binned values
-    m_all_i.setup_binner(n_bins=n_bins)  # TO DO when/where to set up binner?
-    m_all_nmeas.use_binning(m_all_i.binner())
     stats["binned"]["d_max"] = []
     stats["binned"]["d_min"] = []
+    stats["binned"]["n_obs"] = []
     stats["binned"]["n_unique"] = []
     stats["binned"]["completeness"] = []
     stats["binned"]["multiplicity"] = []
@@ -183,29 +181,31 @@ def calc_stats_merged(m_all_i, m_all_nmeas, d_max=0, d_min=0, n_bins=10):
         m_all_i_sel = m_all_i.select(sel)
         m_all_nmeas_sel = m_all_nmeas.select(sel)
         res_low, res_high = m_all_i_sel.d_max_min()
-        n_ref = m_all_i_sel.size()
+        n_unique = m_all_i_sel.size()
         completeness = m_all_i_sel.as_non_anomalous_set().completeness(d_max=res_low)
-        n_ref_nmeas = m_all_nmeas_sel.size()
+        n_obs = int(m_all_nmeas_sel.sum())
+        n_ref_nmeas = int(m_all_nmeas_sel.size())
         multiplicity = m_all_nmeas_sel.mean()
         i_mean =  m_all_i_sel.mean()
         i_sig =  m_all_i_sel.i_over_sig_i()
         stats["binned"]["d_max"].append(round(res_low, 3))
         stats["binned"]["d_min"].append(round(res_high, 3))
-        stats["binned"]["n_unique"].append(n_ref)
-        stats["binned"]["completeness"].append(round(completeness * 100, 1))
-        stats["binned"]["multiplicity"].append(round(multiplicity, 1))
+        stats["binned"]["n_obs"].append(n_obs)
+        stats["binned"]["n_unique"].append(n_unique)
+        stats["binned"]["completeness"].append(round(completeness * 100, 2))
+        stats["binned"]["multiplicity"].append(round(multiplicity, 2))
         stats["binned"]["I"].append(round(i_mean, 2))
         stats["binned"]["IsigI"].append(round(i_sig, 2))
-        # print(f"{res_low:.3f}  {res_high:.3f}  {n_ref}   {completeness * 100:.1f}     {multiplicity:.1f}  {i_mean:.2f}  {i_sig:.2f}")
-    if n_ref != n_ref_nmeas:
-        sys.stderr.write(
-            "WARNING: Numbers of reflections in bins for the intensity Miller"
-            "array and multiplicity Miller array do not equal.\n"
-            "{n_ref} is not {n_ref_nmeas}\n")
+        # print(f"{res_low:.3f}  {res_high:.3f}  {n_unique}   {completeness * 100:.1f}     {multiplicity:.1f}  {i_mean:.2f}  {i_sig:.2f}")
+        if n_unique != n_ref_nmeas:
+            sys.stderr.write(
+                f"WARNING: Numbers of reflections in bins for the intensity Miller"
+                f"array and multiplicity Miller array do not equal.\n"
+                f"{n_ref} is not {n_ref_nmeas} in bin {round(res_low, 3)}-{round(res_high, 3)}\n")
     return stats
 
 
-def calc_stats_compare(m1, m2, d_max, d_min, n_bins):
+def calc_stats_compare(m_all, m1, m2, d_max, d_min, n_bins):
     stats = {"overall": {}, "binned": {}}
     m1 = m1.resolution_filter(d_max=d_max, d_min=d_min)
     m1 = m1.map_to_asu()
@@ -228,8 +228,8 @@ def calc_stats_compare(m1, m2, d_max, d_min, n_bins):
     print(f"CC1/2 = {cc:.3f}\nCC* = {CCstar:.3f}\nRsplit = {rsplit:.3f}")
 
     # binned values
-    m1.setup_binner(n_bins=n_bins)
-    m2.use_binning(m1.binner())
+    m1.use_binning(m_all.binner())
+    m2.use_binning(m_all.binner())
     stats["binned"]["cc"] = []
     stats["binned"]["CCstar"] = []
     stats["binned"]["rsplit"] = []
@@ -269,7 +269,6 @@ def calc_cc_rsplit(half_dataset, spacegroup, cell, d_max=0, d_min=0, n_bins=10):
     half2 = pd.read_csv(
         half_dataset[1], header=None, index_col=False, sep='\s+',
         names=("h", "k", "l", "I", "phase", "sigma(I)", "nmeas"))
-    #print(str(half1["nmeas"][:20])) ERASE
     h1 = flex.int(half1["h"])
     k1 = flex.int(half1["k"])
     l1 = flex.int(half1["l"])
@@ -324,7 +323,6 @@ def calc_cc_rsplit(half_dataset, spacegroup, cell, d_max=0, d_min=0, n_bins=10):
     rsplit = calc_rsplit(m1.data(), m2.data())
     i_sig = m1.i_over_sig_i()
     multiplicity = m1_nmeas.mean()
-    #multiplicity = m1_nmeas.data().mean() NO
     print("\nOverall values:")
     print(f"CC1/2 = {cc:.3f} \nRsplit = {rsplit:.3f} \nI/sigma(I) = {i_sig:.3f}")
     print(f"Multiplicity = {multiplicity:.1f}")
@@ -379,30 +377,83 @@ def stats_to_xml(stats):  #, xmlout="program.xml"):
     return stats_xml
 
 
-def stats_binned_print(stats_binned, half_dataset):
-    stats_print = "%8s%8s%9s%10s%8s%9s%13s"
-    stats_print_header = ["d_max", "d_min", "#unique", "%complet.", "multip.", "<I>", "<I/sigma(I)>"]
-    if half_dataset:
-        stats_print += "%8s%8s%8s"
-        stats_print_header += ["CC(1/2)", "CC*", "Rsplit"]
-    print(stats_print % tuple(stats_print_header))
+def stats_binned_print(stats_binned):  # , half_dataset_available=False):
+    # stats_print = "%9s%9s%11s%8s%9s%9s%9s%9s"
+    header = ["d_max", "d_min", "#obs", "#uniq", "mult.", "%comp", "<I>", "<I/sI>"]
+    stats_print_format_header = '{:>8}{:>8}{:>9}{:>8}{:>8}{:>8}{:>9}{:>8}'
+    stats_print_format_values =  '{:>8.2f}{:>8.2f}{:>9d}{:>8d}{:>8.2f}{:>8.2f}{:>9.1f}{:>8.1f}'
+    # if half_dataset_available:
+    if len(stats_binned) == 11:
+        # stats_print += "%9s%9s%9s"
+        header += ["cc1/2", "cc*", "r_split"]
+        stats_print_format_header += '{:>8}{:>8}{:>8}'
+        stats_print_format_values += '{:>8.3f}{:>8.3f}{:>8.3f}'
+    # print(stats_print % tuple(stats_print_header))
+    print(stats_print_format_header.format(*header))
     for i in range(len(stats_binned["d_max"])):
         values = []
         values.append(stats_binned["d_max"][i])
         values.append(stats_binned["d_min"][i])
+        values.append(stats_binned["n_obs"][i])
         values.append(stats_binned["n_unique"][i])
-        values.append(stats_binned["completeness"][i])
         values.append(stats_binned["multiplicity"][i])
+        values.append(stats_binned["completeness"][i])
         values.append(stats_binned["I"][i])
         values.append(stats_binned["IsigI"][i])
-        if half_dataset:
+        # if half_dataset_available:
+        if len(stats_binned) == 11:
             values.append(stats_binned["cc"][i])
             values.append(stats_binned["CCstar"][i])
             values.append(stats_binned["rsplit"][i])
-        print(stats_print % tuple(values))
+        # print(stats_print % tuple(values))
+        print(stats_print_format_values.format(*values))
     return
 
 
+class MyArgumentParser(argparse.ArgumentParser):
+    """ Helper class for `argparse`
+
+    It adds an attribute `add_argument_with_check` that check if file
+    given in argument exist but not open it.
+    Inspired by `https://codereview.stackexchange.com/questions/28608/
+    checking-if-cli-arguments-are-valid-files-directories-in-python`
+    """
+    def __is_valid_file(self, parser, arg):
+        """Checks if file
+        given in argument `arg` exists but does not open it.
+        If not, abort.
+
+        Args:
+            self
+            parser: parser of `argparse`
+            arg (str): argument of `argparse`
+
+        Returns:
+            str: Name of the checked file
+        """
+        if not os.path.isfile(arg):
+            parser.error('The file {} does not exist!'.format(arg))
+        else:
+            # File exists so return the filename
+            return arg
+
+    def add_argument_with_check(self, *args, **kwargs):
+        """New attribute for `argparse` that checks if file
+        given in argument exist but does not open it.
+        """
+        # Look for your FILE settings
+        # type = lambda x: self.__is_valid_file(self, x) # PEP8 E731
+        def type(x):
+            return self.__is_valid_file(self, x)
+        kwargs['type'] = type
+        self.add_argument(*args, **kwargs)
+
+    def error(self, message):
+        sys.stderr.write('error: %s\n' % message)
+        print('Run `ccp4-python -m import_serial -h` to show the help message.')
+        sys.exit(2)
+
+
 def run():
     from . import __version__
     if not which("f2mtz"):
@@ -410,35 +461,35 @@ def run():
                          "Did you source the paths to CCP4 executables?"
                          "Aborting.\n")
         sys.exit(1)
-    parser = argparse.ArgumentParser(  # TODO requirement
-        description="Covert hkl file(s) from CrystFEL to mtz and calculate statistics"
+    parser = MyArgumentParser(
+        description="Calculate statistics of serial MX data from xia2.ssx or CrystFEL and import them to CCP4"
     )
-    parser.add_argument(
+    parser.add_argument_with_check(
         "--hklin", "--HKLIN",
         metavar=("hklin_file"),
-        help="Specify one merged hkl file from CrystFEL",
+        help="Specify merged mtz file from xia2.ssx or merged hkl file from CrystFEL",
         type=str,
         required=True
     )
     parser.add_argument(
         "--spacegroup",
         type=str,
-        help="Specify space group",
+        help="Space group",
         metavar="spacegroup",
-        required=True
+        # required=True
     )
     parser.add_argument(
         "--cell",
         type=float,
         nargs=6,
-        help="Specify unit cell parameters divided by spaces, e.g. 60 50 40 90 90 90",
+        help="Unit cell parameters divided by spaces, e.g. 60 50 40 90 90 90",
         metavar=("cell_a", "cell_b", "cell_c", "cell_alpha", "cell_beta", "cell_gamma"),
-        required=True
+        # required=True
     )
-    parser.add_argument(
+    parser.add_argument_with_check(
         "--half-dataset",
         metavar=("hkl1", "hkl2"),
-        help="Two half-data-set merge files from CrystFEL (usually .hkl1 and .hkl2)",
+        help="CrystFEL only: two half-data-set merge files (usually .hkl1 and .hkl2)",
         type=str,
         nargs=2,
     )
@@ -450,13 +501,13 @@ def run():
     parser.add_argument(
         "--crystal",
         type=str,
-        help="crystal",
+        help="Crystal name",
         dest='cryst',
     )
     parser.add_argument(
         "--dataset",
         type=str,
-        help="dataset",
+        help="Dataset name",
     )
     parser.add_argument(
         "--dmax", "--lowres",
@@ -479,11 +530,11 @@ def run():
         default=10,
         dest='n_bins',
     )
-    # TO DO: find cell and symmetry from a reference PDB file
     # TO DO: check whether the files exist etc.
     args = parser.parse_args()
 
-    print("Conversion of CrystFEL HKL file to MTZ and import into CCP4")
+    print("")
+    print("Calculate statistics of serial MX data from xia2.ssx or CrystFEL and import them to CCP4")
     print("")
     print("Command line arguments:")
     print(" ".join(sys.argv[1:]))
@@ -492,10 +543,17 @@ def run():
     for arg in vars(args):
         if getattr(args, arg):
             print('  {} {}'.format(arg, getattr(args, arg) or ''))
-    # check whether the files exists TO DO
+
+    hklin = args.hklin
+    if reflection_file_reader.any_reflection_file(hklin).file_type() == 'ccp4_mtz':
+        hklin_format = "dials"
+    elif reflection_file_reader.any_reflection_file(hklin).file_type() == None:
+        hklin_format = "crystfel"
+        spacegroup = None
+        cell_str = None
+
     if args.spacegroup:
         spacegroup = args.spacegroup  # TO DO
-    print("")
     if args.cell:
         cell = args.cell  # TO DO
 
@@ -521,47 +579,92 @@ def run():
     d_max = args.d_max
     d_min = args.d_min
     n_bins = args.n_bins
-    cell_str = str(args.cell[0]) + " " + str(args.cell[1]) + " " + str(args.cell[2]) + " " + str(args.cell[3]) + " " + str(args.cell[4]) + " " + str(args.cell[5])
+    if args.cell:
+        cell_str = str(args.cell[0]) + " " + str(args.cell[1]) + " " + str(args.cell[2]) + " " + str(args.cell[3]) + " " + str(args.cell[4]) + " " + str(args.cell[5])
+    # crystfel: check whether we know spacegroup and cell
+    # dials: direct input arguments in command line for spacegroup and cell have higher priority
 
-    if args.half_dataset:
-        half_dataset = args.half_dataset
-    elif os.path.isfile(args.hklin) and os.path.isfile(args.hklin + "1") and os.path.isfile(args.hklin + "2"):
-        half_dataset = (args.hklin + "1", args.hklin + "2")
-        print(
-            f"Half-dataset files were found automatically and will be used "
-            f"for calculation of statistics: {half_dataset[0]} {half_dataset[1]}")
-    else:
-        half_dataset = None
-    # TO DO: check whether the files exists
-    # TO DO: parse symmetry from .hkl file if not given
+    if hklin_format == "crystfel" and not spacegroup:
+        sys.stderr.write(
+            "ERROR: Space group was not specified but is required for CrystFEL.\n"
+            "       Specify space group explictely or provide reference PDB or mmCIF file.\n"
+            "       Aborting.\n")
+        sys.exit(1)
+    if hklin_format == "crystfel" and not cell_str:
+        sys.stderr.write(
+            "ERROR: Unit cell parameters were not specified but are required for CrystFEL.\n"
+            "       Specify unit cell parameters explictely or provide reference PDB or mmCIF file or .cell file.\n"
+            "       Aborting.\n")
+        sys.exit(1)
+    print("")
+    print("")
+    print("DATA STATISTICS:")
+    print("================")
+    print("")
     try:
-        print("")
-        print("DATA STATISTICS:")
-        print("================")
-        print("")
-        # load data to Miller array
-        cs = crystal.symmetry(
-            unit_cell=uctbx.unit_cell(cell),
-            space_group=sgtbx.space_group_info(spacegroup).group()
-        )
-        m_all_i = get_miller_array(args.hklin, cs, "I", d_max=d_max, d_min=d_min)
-        m_all_nmeas = get_miller_array(args.hklin, cs, "nmeas", d_max=d_max, d_min=d_min)
-        print("Overall values:")
-        print("")
+        m1 = None
+        m2 = None
+        # load data to Miller arrays
+        if hklin_format == "dials":
+            # miller_arrays = reflection_file_reader.any_reflection_file(file_name=hklin).as_miller_arrays()
+            miller_arrays = mtz.object(hklin).as_miller_arrays()
+            m_all_i = None
+            m1_all_nmeas = None
+            m2_all_nmeas = None
+            for i, column in enumerate(miller_arrays):
+                # print(str(column.info().labels))
+                # if column.is_xray_intensity_array() and str(column.info()).split(".mtz:")[1].split(",")[0] == "IMEAN":
+                # elif column.is_xray_intensity_array() and str(column.info()).split(".mtz:")[1].split(",")[0] == "IHALF1":
+                # elif column.is_xray_intensity_array() and str(column.info()).split(".mtz:")[1].split(",")[0] == "IHALF2":
+                if column.is_xray_intensity_array() and column.info().labels == ["IMEAN", "SIGIMEAN"]:
+                    m_all_i = column
+                elif column.is_xray_intensity_array() and column.info().labels == ["IHALF1", "SIGIHALF1"]:
+                    m1 = column
+                elif column.is_xray_intensity_array() and column.info().labels == ["IHALF2", "SIGIHALF2"]:
+                    m2 = column
+                elif column.info().labels == ['N']:
+                    m_all_nmeas = column.as_double()
+        elif hklin_format == "crystfel":
+            if args.half_dataset:
+                half_dataset = args.half_dataset
+            elif os.path.isfile(hklin) and os.path.isfile(hklin + "1") and os.path.isfile(hklin + "2"):
+                half_dataset = (hklin + "1", hklin + "2")
+                print(
+                    f"Half-dataset files were found automatically and will be used "
+                    f"for calculation of statistics: {half_dataset[0]} {half_dataset[1]}")
+            else:
+                half_dataset = None
+            cs = crystal.symmetry(
+                unit_cell=uctbx.unit_cell(cell),
+                space_group=sgtbx.space_group_info(spacegroup).group()
+            )
+            m_all_i = get_miller_array_crystfel(hklin, cs, "I", d_max=d_max, d_min=d_min)
+            m_all_nmeas = get_miller_array_crystfel(hklin, cs, "nmeas", d_max=d_max, d_min=d_min)
+            if half_dataset:
+                m1 = get_miller_array_crystfel(half_dataset[0], cs, "I", d_max=d_max, d_min=d_min)
+                m2 = get_miller_array_crystfel(half_dataset[1], cs, "I", d_max=d_max, d_min=d_min)
+
+        # set d_min, d_max and binning to miller arrays
+        m_all_i = m_all_i.resolution_filter(d_max=d_max, d_min=d_min)
+        m_all_nmeas = m_all_nmeas.resolution_filter(d_max=d_max, d_min=d_min)
+        m_all_i.setup_binner(n_bins=n_bins)
+        m_all_nmeas.use_binning(m_all_i.binner())
+
+        # calculate and print statistics
+        print("Overall values:\n")
         stats_merged = calc_stats_merged(m_all_i, m_all_nmeas, d_max, d_min, n_bins)
-        if half_dataset:
+        if m1 and m2:
             # calculate statistics CC1/2, CC* and Rsplit
-            m1 = get_miller_array(half_dataset[0], cs, "I", d_max=d_max, d_min=d_min)
-            m2 = get_miller_array(half_dataset[1], cs, "I", d_max=d_max, d_min=d_min)
-            stats_compare = calc_stats_compare(m1, m2, d_max, d_min, n_bins)
+            stats_compare = calc_stats_compare(m_all_i, m1, m2, d_max, d_min, n_bins)
             stats_overall = {**stats_merged["overall"], **stats_compare["overall"]}
             stats_binned = {**stats_merged["binned"], **stats_compare["binned"]}
         else:
             stats_overall = {**stats_merged["overall"]}
             stats_binned = {**stats_merged["binned"]}
-
         print("\nBinned values:\n")
-        stats_binned_print(stats_binned, half_dataset)
+        stats_binned_print(stats_binned)
+        
+        # save statistics to files
         stats = {"overall": stats_overall, "binned": stats_binned}
         stats_json = json.dumps(stats, indent=4)
         stats_xml = stats_to_xml(stats)  #, xmlout)
@@ -597,12 +700,21 @@ def run():
     #    # rc = p.returncode
     # remove hkltmp ?
     # print(f"MTZ file created: {hklout}")
-
-    mtz_dataset = m_all_i.as_mtz_dataset(column_root_label="IMEAN")
-    mtz_dataset.add_miller_array(m_all_nmeas, column_root_label="NMEAS")
-    # mtz_dataset.add_miller_array(r_free_flags, column_root_label="FreeR_flag")
-    mtz_dataset.mtz_object().write(hklout)
-    print(f"\nMTZ file created: {hklout}")
-
-    # TO DO - analysis as for 'import_merged' - pointless, aimless, phaser_analysis, ctruncate
+    if hklin_format == "crystfel":
+        mtz_dataset = m_all_i.as_mtz_dataset(column_root_label="IMEAN")
+        mtz_dataset.add_miller_array(m_all_nmeas, column_root_label="NMEAS")
+        # mtz_dataset.add_miller_array(r_free_flags, column_root_label="FreeR_flag")
+        mtz_dataset.mtz_object().write(hklout)
+        print(f"\nMTZ file created: {hklout}")
+        # Clean up a bit
+        files_to_remove = [hklin + "_tmp"]
+        if half_dataset:
+            files_to_remove += [f'{half_dataset[0]}_tmp']
+            files_to_remove += [f'{half_dataset[1]}_tmp']
+        for f in files_to_remove:
+            if os.path.isfile(f):
+                os.remove(f)
+    elif hklin_format == "dials":
+        import shutil
+        shutil.copy2(hklin, hklout)
     return
